""" Compute the All Pair Shortest path and path lengths.

These algorithms work with directed and weighed graphs.

All the algorithms here are deterministic, that means they always compute
the exact path and path length.

The two main approaches to solve the problem are:
    - algorithms that solve the single source shortest problem
    - algorithms base on min-plus matrix multiplication
"""
from heapq import heappush, heappop
from itertools import count

__all__ = [
    "all_pairs_dijkstra",
    "all_pairs_floyd_warshall",
    "cover_labeling",
    "apsp_labeling"
]

def all_pairs_dijkstra(G, weight="weight", *args, **kwargs):
    """Encuentre las rutas y longitudes ponderadas más cortas entre todos los nodos.

    Parameters
    -----------
    G : NetworkX graph
        Grafo dirigido ponderado donde se hara el procesamiento

    weigth : str
        Nombre del atributo donde que contiene el peso de una arista

    Yield
    -----
    (node, (distancia, path)) : (node obj, (dict, dict))
        Cada nodo tienen asociado dos diccionarios. Uno mantiene las distancias(float)
        por cada un de los nodos donde la llave es el nodo correspondiente.
        El segundo contiene el camino, en forma de lista, hacia cada nodo,
        de igual manera el nodo es la llave y cada elemento de lista es nodo del camino (en orden).

    Notes
    -----
    Los pesos de las aristas deben ser numericos.

    Los diccionarios que retorna la funcion solo contiene los nodos accesibles.

    """
    for n in G:
        dist, _ = _single_source_dijkstra(G, n, weight=weight)
        yield (n, dist)


#@distances_decorator
#@run_time_decorator
def all_pairs_floyd_warshall(G, weight="weight",*args, **kwargs):
    """Encuentra todos los caminos más cortos entre todos los pares de vertices
    usando el algoritmo de floyd_warshall

    Parameters
    ----------
    G : NetworkX graph
        Grafo dirigido ponderado donde se hara el procesamiento

    weigth : str
        Nombre del atributo donde que contiene el peso de una arista

    Returns
    -------
    distance : dictionaries
    Diccionario de diccionarios, donde la llave superior corresponde a el nodo i,
    y la inferior al j.

    """
    from collections import defaultdict

    dist = defaultdict(lambda: defaultdict(lambda: float("inf")))
    for u in G:
        dist[u][u] = 0
    for u, v, d in G.edges(data=True):
        e_weight = d.get(weight, 1.0)
        dist[u][v] = min(e_weight, dist[u][v])
    for w in G:
        dist_w = dist[w]  # save recomputation
        for u in G:
            dist_u = dist[u]  # save recomputation
            for v in G:
                d = dist_u[w] + dist_w[v]
                if dist_u[v] > d:
                    dist_u[v] = d
    return dict(dist)


def cover_labeling(G, weigth='weigth', **kwargs):
    """Use Pruned Dijkstra to construct a 2-hop cover labeling from a directed and weighted graph.

    Parameters
    ----------
    G : NetworkX graph
        Directed and weighted graph.

    weigth : str
        Nombre del atributo donde que contiene el peso de una arista

    Returns
    -------
    2-hop cover : dictionaries
        The 'in' key in the dictionary correspond to the coverage for the nodes, taken the in-edges.
        The 'out' key in the dictionary correspond to the coverage for the nodes, taken the out-edges

    Notes
    --------
    To know more about 2-hop cover labeling using pruned algorithms, see the paper:
    https://arxiv.org/abs/1304.4661
    """
    #L = ac.two_hop_cover_dirigido(G)
    cover = {}
    cover['in'] = {}
    cover['out'] = {}
    for v in G:
        cover['in'][v] = dict()
        cover['out'][v] = dict()
    s = sorted(G.degree, key= lambda x: x[1], reverse=True)
    for (k,_) in s:
        cover['out'],_ = _pruned_dijkstra(G, k, cover, edges='out')
        cover['in'],_ = _pruned_dijkstra(G, k, cover, edges='in')
    return cover


def apsp_labeling(G, labeling = None, memory_save=False, **kwargs):
    """Return the APSP distances by yield. The yield statement gets every distance
        from one node to all the others, not one by one distance.

    Parameters
    ----------
    G : NetworkX graph
        Directed and weighted graph.

    labeling : dictionary
        2-Hop Cover Labeling of the graph

    memory_save: boolean
        To reduce the memory consuming by generate 'on the fly' every pair of distances.

    Returns
    -------
    node, disntaces: node, dictionary
        Current node.
        Distances from current node to all the others.
        If memory save flag is true, then every pair of distances is yielded in each iteration,
        and also the value of return is: source node, destination node and distance

    See more
    ---------
        cover_labeling()

    Notes
    --------
    In each yield, the return is a tuple of current node and a
    dictionary of distances from that node to the others.
    """
    #data = ac.calculate_distances(G,l)
    data = dict()
    if labeling is None:
        labeling = cover_labeling(G)
    for i in G.nodes:
        aux_data = dict()
        for j in G:
            aux_data[j] = _query_directes(i, j, labeling)
        if memory_save:
            yield (i,aux_data)
        else:
            data[i] = aux_data
    if not memory_save:
        return dada



def _pruned_dijkstra(G, source, cover, edges='out', weigth='weight'):
    """Pruned Dijkstra, this function uses the coverage to compute the distances from source node
    to all the others. Also it updates the labeling of the nodes that were visited by the source node.

    Parameters
    ----------
    G : NetworkX graph
        Directed and weighted graph.

    source: node
        source node

    cover: dictionary cover
        Coverage o index of the graph

    edges: str
        Type of edges to use; in-edges or out-edges.

    weigth : str
        Atribute name where it´s represented the weigth value.

    Returns
    -------
    coverage updated, distances : dictionary, dictionary
        The coverage updated.
        The distances reached by source node using the cover labeling.
    Notes
    --------
    To know more about 2-hop cover labeling using pruned algorithms, see the paper:
    https://arxiv.org/abs/1304.4661"""
    push = heappush
    pop = heappop
    fringe = []
    dist = {}  # dictionary of final distances
    seen = {}
    c = count()
    adj = G._pred if edges == 'in' else G.adj
    seen[source] = 0
    push(fringe, (0, next(c), source))
    while fringe:
        (d, _, u) = pop(fringe)
        if u in dist:
            continue  # already searched this node.
        dist[u] = d
        if edges == 'out':
            aux_query = _query_directes(source, u, cover)
            if aux_query <= dist[u]:
                dist[u] = aux_query
                continue
        else:
            aux_query = _query_directes(u, source, cover)
            if aux_query <= dist[u]:
                dist[u] = aux_query
                continue
        cover[edges][u].update({source : dist[u]})
        for v, _ in adj[u].items():
            if edges == 'out':
                uv = G.adj[u][v]["weight"]
            else:
                uv = G.adj[v][u]["weight"]
            uv_dist = dist[u] + uv
            if v not in seen or uv_dist < seen[v]:
                seen[v] = uv_dist
                push(fringe, (uv_dist, next(c), v))
    return cover[edges], dist


def _query_directes(source, to, labeling):
    """Search the min distance between two nodes using the cover labeling.

    Parameters
    ----------
    source: node
        Where it from.

    to: node
        Where it to.

    labeling: dictionary cover labeling
        2-Hop Cover Labeling of the graph

    Notes
    ---------
        The function uses a linear serach, but it can gets better performance by using a binary search."""
    labeling_u = dict(labeling['in'][source])
    labeling_v = dict(labeling['out'][to])
    min_distance = float('inf')
    try:
        for k in labeling_u.keys():
            if k in labeling_v:
                aux = (labeling_u[k] + labeling_v[k])
                if aux <= min_distance:
                    min_distance = aux
        return min_distance
    except Exception:
        return float('inf')

def _single_source_dijkstra(G, source, weight):
    """ Encuentra los caminos más cortos entre un vértice dado source
    y todos los demás vértices en el gráfico.

    Parameters
    -----------
    G : NetworkX graph
        Grafo dirigido ponderado donde se hara el procesamiento

    source : NetworkX node
        Nodo de origen


    weigth : str
        Nombre del atributo donde que contiene el peso de una arista

    Returns
    -------
    (distancias, path) : (dict, dict)
        El primer elemento es el diccionario con todas las distancias y
        el segundo elemento es el diccionario de todos los caminos.
    """
    distancias = {}
    padres = {}
    push = heappush
    pop = heappop
    seen = {}
    fringe = []
    seen[source] = 0
    padres[source] = [source]
    c = count()
    push(fringe, (0, next(c), source))
    while fringe:
        (d, _, u) = pop(fringe)
        if u in distancias:
            continue
        distancias[u] = d
        for v in G._adj[u]:
            uv_dist = distancias[u] + G._adj[u][v]["weight"]
            if v not in seen or uv_dist < seen[v]:
                seen[v] = uv_dist
                push(fringe, (uv_dist, next(c), v))
                padres[v] = padres[u] + [v]
    return (distancias,padres)


"""Compute the All Pair Shortest path and path lengths.

These algorithms work with directed and weighed graphs.

All the algorithms here not are deterministic, that means they not always compute
the exact path and path length, but they give a less run time for large graphs.
"""

import numpy as np
import copy
import math
from collections import defaultdict

from disc_python.algoritmos import generic

__all__ = [
    "sketch_apsp",
    "sketches",
    "common_seeds",
    "sketch_shorest_path",
    "all_pairs_ant_colony",
]

def sketch_apsp(G, sketches_labeling = None, memory_save=False, *args, **kwargs):
    """ Find all the shortest distances between all vertices.
    Use the sketches to do the calculation.

    Parameters
    ----------
    G : NetworkX graph
       Directed and weigthed graph

    k : int, (default = 2)
        Amount of times that the generation of a sketch will be executed.

    memory_save: boolean
        To reduce the memory consuming by generate 'on the fly' every pair of distances.

    Returns
    -------
    node, disntaces: node, dictionary
        Current node.
        Distances from current node to all the others.
        If memory save flag is true, then every pair of distances is yielded in each iteration,
        and also the value of return is: source node, destination node and distance

    Notes
    --------
        In each yield, the return is a tuple of current node and a
        dictionary of distances from that node to the others.

    """
    try:
        k = kwargs["k"]
    except KeyError:
        raise KeyError("Give the iterations amount ('k' parameter)")
    if sketches_labeling is None:
        sketches_labeling = sketches(G, k=k)

    data = dict()
    for i in G:
        aux_data = dict()
        for j in G:
            aux_data[j] = sketch_shorest_path(i, j, sketches_labeling)
        if memory_save:
            yield (i,aux_data)
        else:
            data[i] = aux_data
    if not memory_save:
        return dada



def sketches(G, **kwargs):
    """Function that executes k times the generation of a sketch.

    Parameters
    ----------
    G : NetworkX graph
        Directed and weigthed graph.

    k : int
        Number of times that the generation of a sketch will be executed.

    Returns
    -------
    dict : {node_1:[(seed_1,distance_1),...,(seed_r,distance_r) * k],...,node_i}
        sketche with r * k amount of seeds

    Notes
    -----
        The k sketches are combined into one. That is, the combined seeds of k are found in only one.
        For this reason the number of seeds per node is r*k.

    See More
    --------
    _ofline_sample
    """
    try:
        k = kwargs["k"]
    except KeyError:
        raise KeyError("Give the iterations amount ('k' parameter)")
    sketches = {}
    for _ in range(k):
        sketches_samples = _ofline_sample(G)
        for k,v in sketches_samples.items():
            try:
                sketches[k]
            except KeyError:
                sketches[k] = defaultdict(list)
            for kk,vv in v.items():
                sketches[k][kk].extend(vv)
    return sketches

def common_seeds(source, to, sketches):
    """Returns all the seeds (S) that coincide between two nodes, along with their respective distance.
    The distance is the sum of the distance from u to S and from S to v.

    Parameters
    ----------
    source : nodo
        source node.

    to: nodo
        Where it to.

    sketches : dict
        Sketche.

    Returns
    -------
    dict : {seed_i:distance_i,...,seed_S:distance_S}
       Dictionary with all matching seeds.

    Notes
    -----
    If there is a path from u to v then there is also at least one common seed.
    So if it has no seeds in common there is no path and it is represented by float('inf')
    (infinite)
    """

    if source==to:
        return {0:0}
    sketches_u = dict(sketches['in_edge'][source])
    sketches_v = dict(sketches['out_edge'][to])
    intersections_w = {}
    try:
        for k in sketches_u.keys():
            if k in sketches_v:
                distancia = sketches_u[k] + sketches_v[k]
                intersections_w[k] = distancia
        return intersections_w
    except Exception:
        return []

def sketch_shorest_path(source, to, sketches):
    """ Returns the shortest distance between a pair of nodes

    Parameters
    ----------
    source : nodo
        source node.

    to: nodo
        Where it to.

    sketches : dict
        Sketche.

    Returns
    -------
    float : distance
        The minimum distance between all seeds that match between two nodes.
    """

    if source==to:
        return 0.0

    sketche_u = dict(sketches['in'][source])
    sketche_v = dict(sketches['out'][to])
    min_distance = float('inf')
    try:
        for k in sketche_u.keys():
            if k in sketche_v:
                aux = (sketche_u[k] + sketche_v[k])
                if aux <= min_distance:
                    min_distance = aux
        return min_distance
    except Exception:
        return float('inf')

def all_pairs_ant_colony(G, num_ants = 10, max_iter = 1 , *args, **kwargs):
    """ Encuentra todas la ruta más cortas de todos los pares de vertices de
    un grafo dirigido mediante el algoritmo de colonia de hormigas.

    Parameters
    -----------
    G : NetworkX graph
        Grafo dirigido ponderado donde se hara el procesamiento

    num_ants : int (default=10)
        numero de hormigas. El numero de hormigas debe ser igual al numero de nodos.

    max_iter : int (default=100)
        numero maximo de iterciones.

    Returns
    -------
    {
        nodo_origen: {
            nodo_destino: {
                path:[],dist:int
            },
        },
    } : {nodo:{nodo:{path:list,dist:int}}}
        Retorna un diccionario de diccionarios, donde por cada llave de nodo origen
        existen n llaves de nodo destino y cada una de las n llaves destino contienen un diccionario
        donde la primera llave corresponde al path y la seguna a la distancia.

    Notes
    -----
    El numero de hormigas debe ser igual al numero de nodos.
    Al inicializar las variables 'best_path_length' y 'best_path' se hace uso del
    metodo privado _init_pheromones_values, debido a que ambas variables necesitan
    de n^2 iteraciones y se aprovecha el uso de las primeras n iteraciones en está función
    y las siguientes n en el metodo _init_pheromones_values.

    See Also
    --------
    _init_pheromones_values
    _update_pheromones
    """
    i = 0
    best_path = {}
    best_paths = {}
    best_path_length = {}
    pheromones_v = {}
    aux, a, b = _init_pheromones_values(G, best_path, best_path_length)
    for v in G:
        best_path_length[v] = copy.copy(a)
        best_path[v] = copy.copy(b)
        pheromones_v[v] = aux #_init_pheromones_values(G, best_path[v], best_path_length[v])
    while i < max_iter:
        for v in G:
            best_paths[v] = dict()
            for ant in range(num_ants):
                total_edges = []
                best_paths[v][ant] = _path_construction(G, v, ant, pheromones_v[v])
                total_edges.append(best_paths[v][ant]['edges'])
                try:
                    if best_paths[v][ant]['length_path'] <= best_path_length[v][ant]:
                        best_path[v][ant] = {'path':best_paths[v][ant]['path'],"dist":best_paths[v][ant]['length_path']}
                        best_path_length[v][ant] = best_paths[v][ant]['length_path']
                except KeyError as ex:
                    raise(ex)
            _update_pheromones(G, pheromones_v[v], total_edges)
        i += 1
    return best_path_length


def _init_pheromones_values(G, best_path={}, best_path_length={}):
    """ Inicializa una lista de adyacencia de feromonas

    Parameters
    -----------
    G : NetworkX graph
        Grafo dirigido ponderado donde se hara el procesamiento

    best_path : dict, optional(default={})
        Diccionario de los mejores caminos encontrados.

    best_path_length : int, optional(default={})
        Diccionario de las mejores longitudes encontradas por cada camino.

    Returns
    --------
    {node:{node_neighbor:{pheromone:float}}} : {node:{node:{srt:float}}}
        Lista de adyacencia que contiene el nivel de feromona que tiene cada arista.

    Notes
    -----
    Los parametros best_path, best_path_length son uso exclusivo de la funcion all_pairs_ant_colony,
    por lo que si se desea utilizar esta funcion en algun otro sitio es recomendable no pasarle
    como parametros nada en estas dos variables.
    El nivel de feromona se establece con la formula 1/grado_de_salida_del_vertice(n)

    See Also
    ---------
    all_pairs_ant_colony
    """
    adj_pheromones = {}
    for n in G:
        adj_pheromones[n] = dict()
        best_path[n] = []
        best_path_length[n] = float('inf')
        for neighbor in G.adj[n]:
            adj_pheromones[n][neighbor] = {"pheromone":1/G.out_degree(neighbor) if G.out_degree(neighbor) != 0 else 0 }
    return (adj_pheromones, copy.copy(best_path_length), copy.copy(best_path))

def _path_construction(G, source, target, pheromones):
    """ Construye un camino simple entre dos nodos, atraves de una caminata aleatoria
    con interaccion entre la cantidad de pheromonas que existe en las aristas.

    Parameter
    ----------
    G : NetworkX graph
        Grafo dirigido ponderado donde se hara el procesamiento

    source : Network nodo
        Nodo de origen

    target : Network nodo
        Nodo de destino

    pheromones : dict
        Lista de adyacencia de las feromonas

    Returns
    -------
    {path:[],edges:[],length_path:float} : dict()
        Diccionario con las claves:
            path        : lista de los nodos que contiene el camino
            edges       : lista de tuplas, cada tupla es una arista del tipo e=(v1,v2)
            length_path : Distancia total del camino

    Notes
    -----
    Es posible que en el recorrido no se llegue al destino, si esto ocurre,
    entonces, la distacia es infinito y el path son los nodos que se recorrieron.
    La distancia del path se calcula en _eval_path.
    Está version es una version con interaccion, lo que significa que las hormigas
    utilizan informacion de las feromonas de otra colonia, junto con la suya propia
    para encontrar el camino más corto.

    See Also
    --------
    _eval_path
    """
    p = [source]
    i = 0
    v = G.adj[source]
    edges = []
    visited = [source]
    try:
        G.adj[target]
    except KeyError:
        raise KeyError("No hay camino hacia {}".format(target))
    while list(v) != [] and target not in p:
        demoninator = 0
        for neighbor in v:
            demoninator += pheromones[p[i]][neighbor]['pheromone']
        for neighbor in v:
            prob = np.random.uniform(low=0.0, high=1.0, size=None)
            if (pheromones[p[i]][neighbor]['pheromone']/demoninator) >= prob:
                p.append(neighbor)
                v = set(G.adj[neighbor]).difference(visited)
                edges.append((p[i], neighbor))
                visited.append(neighbor)
                i += 1
                break
    return {"path": p, "edges":edges, "length_path":_eval_path(G, edges, target)}


def _update_pheromones(G, pheromones, edges, p=0.5):
    """Actualiza una lista de adyacencia (feromonas)

    Parameters
    -----------
    G : NetworkX graph
        Grafo dirigido ponderado donde se hara el procesamiento

    pheromones : dict , (paso por valor).
        Lista de adyacencia de feromonas a actualizar.

    edges : list of tuples
        Lista de tuplas, donde cada tupla es una arista. Este parametro debe contener todas las
        aristas de todos los caminos encontrados.

    p : float (default=0.5)
        Tasa de evaporación de feromonas

    Returns
    -------
    None

    Notes
    -----
    No retorna nada debido a que la lista de adyacencia de feromonas se pasa por valor y no por
    referencia.
    """
    for n, neighbors in G.adjacency():
        for neighbor in neighbors:
            if (n, neighbor) in edges:
                pheromones[n][neighbor]['pheromone'] = min(1-p*pheromones[n][neighbor]['pheromone']+p, (G.number_of_nodes()**2-1)/G.number_of_nodes()**2)
            else:
                pheromones[n][neighbor]['pheromone'] = max(1-p*pheromones[n][neighbor]['pheromone']+p, 1/G.number_of_nodes()**2)

def _eval_path(G, edges, target):
    """Calcula la distancia total de un camino

    Parameters
    ----------
    G : NetworkX graph
        Grafo dirigido ponderado donde se hara el procesamiento.

    edges : list of tuples
        Lista de tuplas, donde cada tupla es una arista del camino.

    target : NetworkX node
        Nodo de destino, a donde se supone debe terminar el camino.

    Returns
    -------
    length : float
        Retorna la distancia total del camino

    Notes
    -----
    En caso de que la lista de aristas no conduzcan a el nodo destino,
    se retornara infinito.
    """
    length = 0
    if edges == [] or edges[-1][1] is not target:
        return float('inf')
    for f,t in edges:
        length += G.adj[f][t]['weight']
    return length


def _ofline_sample(G):
    """ Function that returns a sketch with r nearest number of seeds.
        r is a sample that depends on the number of nodes, of degree > 2, of the graph.
        Specifically for this version, a base 2 logarithmic sample is used.

    Parameters
    ----------
    G : NetworkX graph
        Directed and weigthed graph

    Returns
    -------
    dict : {node_1:[(seed_1,distance_1),...,(seed_r,distance_r)],...,node_i}
        sketch with r seeds

    Notes:
    -------
        It is possible that not all nodes have r amounts of seeds,
        this can happen if the node has no path to any seed.
        However this is not an error, because if that were to happen it means
        that the node is not connected to the graph or that it has no output edges, so
        there cannot be a path from that node to the others (but to it if it only has input edges).

        The function is only adapted for weighted directed graphs. However it is easily
        generalize it for undirected (unweighted) graphs.

    See More
    --------
        http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.167.9222&rep=rep1&type=pdf
    """
    r = math.floor(math.log2(G.number_of_nodes()))
    degrees = dict(G.degree())
    nodes_v = list(degrees.keys())
    sum_d = sum(degrees.values())
    probs = [x / sum_d for x in degrees.values()]
    sets = [set(np.random.choice(nodes_v, size=2**i, p=probs,replace=False)) for i in range(r)]
    sketches = {}
    directed = [("in",True), ("out",False)]
    for s in sets:
        for d,b in directed:
            path = generic.dijkstra_multisource(G, list(s), in_edges=b)
            try:
                sketches[d]
            except KeyError:
                sketches[d] = defaultdict(list)
            for k,v in path.items():
                sketches[d][k].append(v)
    return sketches




import pathlib
import subprocess
import sys
from setuptools import find_packages, setup
from setuptools.command.install import install as _install


HERE = pathlib.Path(__file__).parent

VERSION = '1.0.1'
PACKAGE_NAME = 'disc_python'
AUTHOR = 'Esteban Mendiola Tellez'
AUTHOR_EMAIL = 'mendiola_esteban@outlook.com'
URL = 'https://gitlab.com/TebanMT'

LICENSE = 'MIT'
DESCRIPTION = 'Library to compute conceptual distance by using DIS-C method'
LONG_DESCRIPTION = (HERE / "README.md").read_text(encoding='utf-8')
LONG_DESC_TYPE = 'text/markdown'


INSTALL_REQUIRES = [
    'networkx',
    'numpy',
    'nltk',
    ]

class Install(_install):
    def run(self):
        _install.do_egg_install(self)
        import nltk
        nltk.download("wordnet")

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESC_TYPE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    #cmdclass={'install': Install},
    install_requires=INSTALL_REQUIRES,
    setup_requires=INSTALL_REQUIRES,
    license=LICENSE,
    packages=find_packages(),
    include_package_data=True
)



import sys
import networkx as nx
from nltk.corpus import wordnet as wn


class WordNet():
    """
    Clase para el corpus WordNet

    Contienen toda la logica para acceder a los synsets de WordNet

    Los metodos 'publicos' de la clase retornan siempre un grafo NetworkX
    para su manipulacion.

    Cada vertice en el grafo es un lemma_names() y cada arista es el tipo de relacion
    que los une.
    """

    def __init__(self):
        self.use_type = {
            "HYPERNYMS": True, "HYPONYMS": True,"INSTANCE_HYPERNYMS": True,
            "INSTANCE_HYPONYMS": True, "MEMBER_HOLONYMS": True,
            "MEMBER_MERONYMS": True, "PART_HOLONYMS": True, "PART_MERONYMS": True
        }

    def _get_synsets(self, word):
        """Obtiene todos los synsets relacionados con un terminos

        Parameters
        ----------
        word : string
            El termino al que le desea buscar sus synsets

        Returns
        -------
        synsets : list
            Una lista con todos los synsets
        """
        synsets = wn.synsets(word)
        return synsets

    def _get_relations(self, word):
        """Obtiene en un diccionario los terminos relacionados con un termino de entrada.
        Cada termino relacionado pertenece a un tipo de relacion descrita por una de
        las llaves del diccionario de retorno.

        Parameters
        ----------
        word : string
            El termino al que se le desea obtner sus relaciones

        Returns
        -------
        relaciones : dict
            Un diccionario que contiene todos los terminos relacionados a un termino
            separados por el tipo de relacion.

        Notes
        -----
        Los tipos de relaciones incluidas en el diccionario de salida son todos aquellos
        que se encuentran con la bandera True en el atributo de clase use_type
        """
        realaciones = {}
        synsets = self._get_synsets(word)
        for k,v in self.use_type.items():
            if(v):
                realaciones[k] = []
                for synset in synsets:
                    if(synset.pos() == wn.NOUN):
                        for tss in getattr(synset, k.casefold())():
                            lemmas_name = tss.lemma_names()
                            realaciones[k] = realaciones[k] + lemmas_name
        return realaciones

    def get_relations_like_graph(self, word):
        """Retorna las relaciones de un termino a manera de grafo

        Parameters
        ----------
        word : str
            Termino de entrada

        Returns
        -------
        graph : NetworkX graph
            Grafo con las relaciones de un termino. Cada vertice del grafo
            es un termino
        """
        relaciones = self._get_relations(word)
        graph = nx.DiGraph()
        if relaciones == {}:
            print("La palabra {} No existe en el Corpus".format(word))
            return graph
        for k,v in relaciones.items():
            for r in v:
                if graph.has_node(r) == False:
                    graph.add_edge(word, r, tipo=k)
        return graph

    def to_graph_from_list_words(self, words):
        """Intenta retornar un grafo conectado que contenga las palabras
        que se le pasen como parametro.

        Parameters
        ----------
        words : list
            Lista de terminos

        Returns
        -------
        graph : NetworkX graph
            Grafo que contiene los terminos relacionados con cada termino en la lista de entrada.

        Notes
        -----
        El grafo de retorno es conectado siempre y cuando la profundidad de la busqueda
        logre conectar en algun momento todos los terminos.

        See Also
        --------
        _to_graph
        """
        g = nx.DiGraph()
        for w in words:
            self._to_graph(g,None,w,None)
        return g

    def _to_graph(self, graph, from_word, to_word, type_relation, depth=1, max_depth=2):
        """Funcion recursiva que busca concetar dos terminos mediante sus relaciones

        Parameters
        ----------
        graph : NetworkX graph
            Grafo de entrada. Puede ser un grafo vacio o que ya contenga datos

        from_word : string
            Primer termino que se busca conectar con el segundo termino (to_word).
            Si se le proporciona como entrada el valor de None significa que no se
            intentara concetar dos terminos y solo se buscaran las relaciones del
            segundo termino.

        to_word : string
            Termino con el que se busca conectar el primer termino

        type_relation : string
            Tipo de relacion. Este parametro se utiliza en la funcion recursiva por lo cual
            no se debe proporcionar cuando se llame la funcion en primera instancia.

        depth : int, optional (default = 1)
            Profundidad a la que se encuentra la busqueda. Tener en cuenta que si se modifica
            el default se debe asegurar que max_depth tambien se modifique para que se cumpla depth <= max_depth

        max_depth : int, optional (default = 2)
            Profundidad maxima a la que se desea buscar las relaciones de los terminos

        Returns
        -------
        graph : NetworkX graph
            Grafo con todos los terminos relacionados

        Notes
        -----
        Que se conecten los terminos de entrada en un grafo depende de la profundidad
        maxima a la que se desea buscar, si en el nivel maximo de profundidad no encuentra el
        termino destino, entonces no se logro conctar los terminos. Cada nivel de profundidad
        son las relaciones conectadas a un termino. Ejemplo:

                O
               / \   <-- profundidad 1
               1  2
              / \ /\ <-- profundidad 2
                 .
                 .
                 .

        """
        to_exist = graph.has_node(to_word)
        if from_word is not None:
            graph.add_edge(from_word, to_word, tipo=type_relation)
        if depth <= max_depth and not to_exist:
            relaciones = self._get_relations(to_word)
            for k,v in relaciones.items():
                for element in v:
                    self._to_graph(graph, to_word, element, k, depth+1, max_depth)
        return graph

    def to_graph_connect_words(self, source, target, inter_word=None, max_depth=50):
        """Intenta retornar un grafo que conecta dos terminos,
        pasando por un termino intermedio, a traves de una busqueda en profundidad

        Parameters
        ----------
        source : str
            Termino fuente, donde se iniciara la busqueda

        target: str
            Termino destino, donde se intentara terminar la busqueda

        inter_word : str
            Termino intermedio, el termino que por el cual debe pasar la busqueda. Si no se
            proporciona ninguna palabra intermedia, entonces solo se intenta conectar el source
            con el target

        max_depth: int
            Profundidad maxima a la que se desea buscar las relaciones de los terminos

        Returns
        -------
        graph : NetworkX graph
            Grafo

        Notes
        -----
        El mecanismo logico detras de la funcion es concetar primero source con inter_word
        y despues target con inter_word, así source y target estaran concetados. Depende de la
        profundidad maxima de busqueda.

        See Also
        --------
        _dfs
        """
        graph = nx.DiGraph()
        if inter_word:
            self._dfs(graph, source, inter_word, max_depth=max_depth)
            self._dfs(graph, target, inter_word, max_depth=max_depth)
        else:
            self._dfs(graph, source, target, max_depth=max_depth)
        return graph
        #return self._dfs(graph, source, inter_word)

    def _dfs(self, graph, source, target, depth=1, max_depth=20):
        """Usa el algoritmo DFS para conectar dos terminos

        Parameters
        ----------
        graph : NetworkX graph
            Grafo

        source : str
            Termino fuente, de donde comenzara la busqueda

        target : str
            Termino destino, donde terminara la busqueda

        depth : int, optional (default = 1)
            Profundidad a la que se encuentra la busqueda. Tener en cuenta que si se modifica
            el default se debe asegurar que max_depth tambien se modifique para que se cumpla depth <= max_depth

        max_depth : int, optional (default = 20)
            Profundidad maxima a la que se desea buscar las relaciones de los terminos

        Notes
        -----
        Que se conecten los terminos de entrada en un grafo depende de la profundidad
        maxima a la que se desea buscar, si en el nivel maximo de profundidad no encuentra el
        termino destino, entonces no se logro conctar los terminos. Cada nivel de profundidad
        son las relaciones conectadas a un termino. Ejemplo:

                O
               / \   <-- profundidad 1
               1  2
              / \ /\ <-- profundidad 2
                 .
                 .
                 .
        """
        relaciones = self._get_relations(source)
        if relaciones == {}:
            print("La palabra {} No existe en el Corpus".format(source))
            return graph
        if depth > max_depth:
            return graph
        for k,v in relaciones.items():
            if target in v:
                graph.add_edge(source, target, tipo=k)
                return graph
            #Buesqueda en profundidad del nodo target
            for element in v:
                if graph.has_node(element):
                    graph.add_edge(source, element, tipo=k)
                else:
                    graph.add_edge(source, element, tipo=k)
                    self._dfs(graph, element, target, depth=depth+1, max_depth=max_depth)
        return graph